from typing import Final

DEFAULT_LAT_LON_FOR_SOLPOS: Final = (36.7128, -122.186)

DEFAULT_TITLE: Final = (
    "MBARI MARS Cabled Observatory   Monterey Bay, California, USA"
    f"   {DEFAULT_LAT_LON_FOR_SOLPOS[0]} °N, {DEFAULT_LAT_LON_FOR_SOLPOS[1]} °W"
)

DEFAULT_YLIM: Final = (10, 100000)

DEFAULT_CMLIM: Final = (32, 108)

DEFAULT_DPI: Final = 200
